# SPDX-FileCopyrightText: 2020 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: CC0-1.0

query stationsByName($name: String!) {
  stations: quays(name: $name) {
    id
    name
    lat: latitude
    lon: longitude
    timezone: timeZone
    parentStation: stopPlace {
      id
      lat: latitude
      lon: longitude
      timezone: timeZone
    }
  }
}
