# SPDX-FileCopyrightText: 2020 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: CC0-1.0

query stationsByName($name: String!) {
  stations: stops(name: $name) {
    id: gtfsId
    name
    lat
    lon
    parentStation {
      id: gtfsId
      name
      lat
      lon
      timezone
    }
    timezone
  }
}
