# SPDX-FileCopyrightText: 2020-2021 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: CC0-1.0

query stationsByName($name: String!) {
    stations: stops(name: $name) {
        id: gtfsId
        name
        lat
        lon
        vehicleMode
        parentStation {
            id: gtfsId
            name
            lat
            lon
        }
    }
}
