/*
    SPDX-FileCopyrightText: 2019 Volker Krause <vkrause@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "disruption.h"

using namespace KPublicTransport;

#include "moc_disruption.cpp"
